from typing import Optional

from . import ExtensionMixin
from ..session import FlarumUser

from ..flarum.core.filters import Filter


AUTHOR = 'skevo'
NAME = 'absolutely-all'
ID = f"{AUTHOR}-{NAME}"


SOFT_DEPENDENCIES = []
HARD_DEPENCENDIES = []


class AbsolutelyAllFlarumUserMixin(FlarumUser):
    def absolutely_all_users(self, filter: Optional[Filter]=None):
        """
            A generator that yields `Users` from entire forum, until there are `None` left. `Filter` compatible.
        """

        if isinstance(filter, Filter):
            _filter = filter

        else:
            _filter = Filter()
    
        _filter.page = 0
        users = self.all_users(_filter)

        while True:
            yield users

            users = self.all_users(_filter)
            _filter.page += 1

            if not users.next_link:
                break


    def absolutely_all_posts(self, filter: Optional[Filter]=None):
        """
            A generator that yields `Posts` from entire forum, until there are `None` left. `Filter` compatible.
        """

        if isinstance(filter, Filter):
            _filter = filter

        else:
            _filter = Filter()
    
        _filter.page = 0
        posts = self.all_posts(_filter)

        while True:
            yield posts

            posts = self.all_posts(_filter)
            _filter.page += 1

            if not posts.next_link:
                break


    def absolutely_all_discussions(self, filter: Optional[Filter]=None):
        """
            A generator that yields `Discussions` from entire forum, until there are `None` left. `Filter` compatible.
        """

        if isinstance(filter, Filter):
            _filter = filter

        else:
            _filter = Filter()
    
        _filter.page = 0
        discussions = self.all_discussions(_filter)

        while True:
            yield discussions

            discussions = self.all_discussions(_filter)
            _filter.page += 1

            if not discussions.next_link:
                break


    def absolutely_all_notifications(self, filter: Optional[Filter]=None):
        """
            A generator that yields all of your `Notifications`, until there are `None` left. `Filter` compatible.
        """

        if isinstance(filter, Filter):
            _filter = filter

        else:
            _filter = Filter()
    
        _filter.page = 0
        discussions = self.get_notifications(_filter)

        while True:
            yield discussions

            discussions = self.get_notifications(_filter)
            _filter.page += 1

            if not discussions.next_link:
                break


class AbsolutelyAllExtension(ExtensionMixin):
    def __init__(self):
        self.name = NAME
        self.author = AUTHOR
        self.id = ID

    def get_dependencies(self):
        return {
            "soft": SOFT_DEPENDENCIES,
            "hard": HARD_DEPENCENDIES
        }


    def mixin(self):
        super().mixin(self, FlarumUser, AbsolutelyAllFlarumUserMixin)



from typing import Optional, Union

from datetime import datetime


def flarum_to_datetime(flarum_dt: Optional[Union[str, datetime]]=None):
    """
        Converts Flarum's datetime string to Python's datetime object.
        Doesn't convert if the parameter is already a datetime object.
        
        Flarum's datetime format is `%Y-%m-%dT%H:%M:%S%z`
    """

    if not flarum_dt:
        return None

    elif isinstance(flarum_dt, datetime):
        return flarum_dt

    else:
        strpped = datetime.strptime(flarum_dt, r'%Y-%m-%dT%H:%M:%S%z') # type: datetime

        return strpped


def datetime_to_flarum(dt: Optional[Union[datetime, str]]=None):
    """
        Converts Python's datetime object to Flarum's datetime string.
        Doesn't convert if the parameter is already a string.
        
        Flarum's datetime format is `%Y-%m-%dT%H:%M:%S%z`
    """

    if not dt:
        return None

    elif isinstance(dt, str):
        return dt

    else:
        strfed = dt.strftime(r'%Y-%m-%dT%H:%M:%S%z') # type: str

        return strfed


"""
    Custom types for pyFlarum.

    ### Types:
    - `AnyUser` - `User`, `UserFromBulk`, `UserFromNotification`
    - `AnyDiscussion` - `Discussion`, `DiscussionFromBulk`, `DiscussionFromNotification`
    - `AnyPost` - `Post`, `PostFromBulk`, `PostFromNotification`
"""

from typing import Union


from .flarum.core.users import User, UserFromBulk, UserFromNotification
from .flarum.core.discussions import Discussion, DiscussionFromBulk, DiscussionFromNotification
from .flarum.core.posts import Post, PostFromBulk, PostFromNotification


AnyUser = Union[User, UserFromBulk, UserFromNotification]
AnyDiscussion = Union[Discussion, DiscussionFromBulk, DiscussionFromNotification]
AnyPost = Union[Post, PostFromBulk, PostFromNotification]



from typing import Any, Callable, Optional

import time

from . import ExtensionMixin
from ..session import FlarumUser
from ..flarum.core.notifications import Notification


AUTHOR = 'skevo'
NAME = 'watch'
ID = f"{AUTHOR}-{NAME}"


SOFT_DEPENDENCIES = []
HARD_DEPENCENDIES = []


class WatchFlarumUserMixin(FlarumUser):
    def watch_notifications(self, on_notification: Callable[[Notification], Any], interval: Optional[float]=10, auto_mark_as_read: bool=True, **kwargs):
        while True:
            all_notifications = self.get_notifications(**kwargs)

            for notification in all_notifications:
                if not notification.isRead:
                    on_notification(notification)

                    if auto_mark_as_read:
                        notification.mark_as_read()

            if interval:
                time.sleep(interval)


class WatchExtension(ExtensionMixin, WatchFlarumUserMixin):
    def __init__(self):
        self.name = NAME
        self.author = AUTHOR
        self.id = ID

    def get_dependencies(self):
        return {
            "soft": SOFT_DEPENDENCIES,
            "hard": HARD_DEPENCENDIES
        }


    def mixin(self):
        super().mixin(self, FlarumUser, WatchFlarumUserMixin)



from typing import List, Optional, TYPE_CHECKING

# Avoid my greatest enemy in Python: circular import:
if TYPE_CHECKING:
    from ...session import FlarumUser

from datetime import datetime

from ...error_handler import parse_request
from ...datetime_conversions import flarum_to_datetime


class Users(dict):
    """
        A data of multiple users fetched from the API.
    """

    def __init__(self, user: 'FlarumUser', _fetched_data: dict):
        self.user = user

        super().__init__(_fetched_data)


    def __iter__(self):
        return iter(self.get_users())


    @property
    def links(self) -> dict:
        return self.get("links", {})


    @property
    def first_link(self) -> Optional[str]:
        return self.links.get("first", None)


    @property
    def previous_link(self) -> Optional[str]:
        return self.links.get("prev", None)


    @property
    def next_link(self) -> Optional[str]:
        return self.links.get("next", None)


    @property
    def data(self) -> List[dict]:
        return self.get("data", [{}])


    @property
    def included(self) -> List[dict]:
        return self.get("included", [{}])


    def get_users(self):
        """
            All users from the `Users` object.
        """

        all_users = [] # type: List[UserFromBulk]

        for raw_user in self.data:
            if raw_user.get("type", None) == 'users':
                user = UserFromBulk(user=self.user, _fetched_data=dict(data=raw_user, _parent_included=self.included))
                all_users.append(user)

        return all_users


class UserFromNotification(dict):
    """
        An user from `BaseNotification`
    """

    def __init__(self, user: 'FlarumUser', _fetched_data: dict):
        self.user = user
        super().__init__(_fetched_data)


    @property
    def data(self) -> dict:
        return self.get("data", {})

    @property
    def type(self) -> Optional[str]:
        return self.data.get("type", None)


    @property
    def id(self) -> Optional[int]:
        return self.data.get("id", None)


    @property
    def attributes(self) -> dict:
        return self.data.get("attributes", {})


    @property
    def username(self) -> Optional[str]:
        return self.attributes.get("username", None)


    @property
    def email(self) -> Optional[str]:
        return self.attributes.get("email", None)


    @property
    def isEmailConfirmed(self) -> bool:
        return self.attributes.get("isEmailConfirmed", False)


    @property
    def displayName(self) -> Optional[str]:
        return self.attributes.get("displayName", None)


    @property
    def avatarUrl(self) -> Optional[str]:
        return self.attributes.get("avatarUrl", None)


    @property
    def slug(self) -> Optional[str]:
        return self.attributes.get("slug", None)


class UserFromBulk(UserFromNotification):
    """
        An user from `Users`.
    """

    def __init__(self, user: 'FlarumUser', _fetched_data: dict):
        self.user = user
        super().__init__(user=self.user, _fetched_data=_fetched_data)


    @property
    def joinTime(self) -> Optional[datetime]:
        raw = self.attributes.get("joinTime", None)

        return flarum_to_datetime(raw)


    @property
    def discussionCount(self) -> Optional[int]:
        return self.attributes.get("discussionCount", None)


    @property
    def commentCount(self) -> Optional[int]:
        return self.attributes.get("commentCount", None)


    @property
    def canEdit(self) -> bool:
        return self.attributes.get("canEdit", False)


    @property
    def canEditCredentials(self) -> bool:
        return self.attributes.get("canEditCredentials", False)


    @property
    def canEditGroups(self) -> bool:
        return self.attributes.get("canEditGroups", False)


    @property
    def canDelete(self) -> bool:
        return self.attributes.get("canDelete", False)


    @property
    def relationships(self) -> dict:
        return self.data.get("relationships", {})



class User(UserFromBulk):
    """
        An user that was fetched from the API.
    """

    def __init__(self, user: 'FlarumUser', _fetched_data: dict):
        self.user = user
        super().__init__(user=self.user, _fetched_data=_fetched_data)


    @property
    def foo(self):
        return 'bar'



class MyUser(User):
    """
        Your user, contains full user data.
    """

    def __init__(self, user: 'FlarumUser', _fetched_data: dict):
        self.user = user

        super().__init__(user=self.user, _fetched_data=_fetched_data)



from typing import TYPE_CHECKING, Optional, List

# Avoid my greatest enemy in Python: circular import:
if TYPE_CHECKING:
    from ...session import FlarumUser

from datetime import datetime

from ...flarum.core.users import User

from ...error_handler import FlarumError, parse_request
from ...datetime_conversions import flarum_to_datetime


class PreparedDiscussion(dict):
    def __init__(self, user: 'FlarumUser', title: Optional[str]=None, content: Optional[str]=None):
        self.user = user
        self.title = title
        self.content = content

        super().__init__()


    @property
    def to_dict(self):
        data = {
            "data": {
                "type": "discussions",
                "attributes": {
                    "title": self.title,
                    "content": self.content
                }
            }
        }

        return data


    def post(self):
        """
            Posts/creates the discussion. Raises `FlarumError` or returns `False` if it failed, otherwise the new `Discussion` is returned.
        """

        if not isinstance(self.title, str) or not isinstance(self.content, str):
            raise TypeError(f"Both `title` and `content` parameters must be a `str`.")

        raw = self.user.session.post(self.user.api_urls['discussions'], json=self.to_dict)
        json = parse_request(raw)

        return Discussion(user=self.user, _fetched_data=json)
    create = post


class Discussions(dict):
    """
        A data of multiple discussions fetched from the API.
    """

    def __init__(self, user: 'FlarumUser', _fetched_data: dict):
        self.user = user

        super().__init__(_fetched_data)


    def __iter__(self):
        return iter(self.get_discussions())


    @property
    def links(self) -> dict:
        return self.get("links", {})


    @property
    def first_link(self) -> Optional[str]:
        return self.links.get("first", None)


    @property
    def previous_link(self) -> Optional[str]:
        return self.links.get("prev", None)


    @property
    def next_link(self) -> Optional[str]:
        return self.links.get("next", None)


    @property
    def data(self) -> List[dict]:
        return self.get("data", [{}])


    @property
    def included(self) -> List[dict]:
        return self.get("included", [{}])


    def get_discussions(self):
        """
            All discussions from the `Discussions` object.
        """

        all_discussions = [] # type: List[DiscussionFromBulk]

        for raw_discussion in self.data:
            if raw_discussion.get("type", None) == 'discussions':
                discussion = DiscussionFromBulk(user=self.user, _fetched_data=dict(data=raw_discussion, _parent_included=self.included))
                all_discussions.append(discussion)

        return all_discussions


class DiscussionFromNotification(dict):
    """
        A discussion from `BaseNotification`
    """

    def __init__(self, user: 'FlarumUser', _fetched_data: dict):
        self.user = user
        super().__init__(_fetched_data)


    def get_full_data(self):
        raw = self.user.session.get(f"{self.user.api_urls['discussions']}/{self.id}")
        json = parse_request(raw)

        return Discussion(user=self.user, _fetched_data=json)


    @property
    def data(self) -> dict:
        return self.get("data", {})


    @property
    def type(self) -> Optional[str]:
        return self.data.get("type", None)


    @property
    def id(self) -> Optional[int]:
        return self.data.get("id", None)


    @property
    def attributes(self) -> dict:
        return self.data.get("attributes", {})


    @property
    def title(self) -> Optional[str]:
        return self.attributes.get("title", None)


    @property
    def slug(self) -> Optional[str]:
        return self.attributes.get("slug", None)


    def __restore_or_hide(self, hide: bool):
        patch_data = {
            "data": {
                "type": "discussions",
                "id": self.id,
                "attributes": {
                    "isHidden": hide
                }
            }
        }

        raw = self.user.session.patch(f"{self.user.api_urls['discussions']}/{self.id}", json=patch_data)
        parse_request(raw)

        return True


    def hide(self):
        """
            Hides the discussion. Raises `FlarumError` or returns `False` if it failed, otherwise `True` is returned.
        """

        return self.__restore_or_hide(hide=True)


    def restore(self):
        """
            Restores the discussion (unhides). Raises `FlarumError` or returns `False` if it failed, otherwise `True` is returned.
        """

        return self.__restore_or_hide(hide=False)
    unhide = restore


    def delete(self):
        """
            Deletes a discussion forever - this action is irreversible!
        """

        raw = self.user.session.delete(f"{self.user.api_urls['discussions']}/{self.id}")
        parse_request(raw)

        return True



# Circular:
from ...flarum.core.posts import PostFromDiscussion
class DiscussionFromBulk(DiscussionFromNotification):
    """
        A discussion from `Discussions`.
    """

    def __init__(self, user: 'FlarumUser', _fetched_data: dict):
        self.user = user
        super().__init__(user=self.user, _fetched_data=_fetched_data)


    @property
    def url(self):
        slug = self.slug

        if slug:
            return f"{self.user.forum_url}/d/{slug}"

        else:
            return f"{self.user.forum_url}/d/{self.id}"


    @property
    def commentCount(self) -> Optional[str]:
        return self.attributes.get("commentCount", None)


    @property
    def participantCount(self) -> Optional[str]:
        return self.attributes.get("participantCount", None)


    @property
    def createdAt(self) -> Optional[datetime]:
        raw = self.attributes.get("createdAt", None)

        return flarum_to_datetime(raw)


    @property
    def lastPostedAt(self) -> Optional[datetime]:
        raw = self.attributes.get("lastPostedAt", None)

        return flarum_to_datetime(raw)


    @property
    def lastPostNumber(self) -> Optional[int]:
        return self.attributes.get("lastPostNumber", None)


    @property
    def lastReadPostNumber(self) -> Optional[int]:
        return self.attributes.get("lastReadPostNumber", None)


    @property
    def canReply(self) -> bool:
        return self.attributes.get("canReply", False)


    @property
    def canRename(self) -> bool:
        return self.attributes.get("canRename", False)


    @property
    def canDelete(self) -> bool:
        return self.attributes.get("canDelete", False)


    @property
    def canHide(self) -> bool:
        return self.attributes.get("canHide", False)


    @property
    def lastReadAt(self) -> Optional[datetime]:
        raw = self.attributes.get("lastReadAt", None)

        return flarum_to_datetime(raw)


    @property
    def isHidden(self) -> bool:
        return self.attributes.get("isHidden", False)


    @property
    def subscription(self) -> Optional[str]:
        return self.attributes.get("subscription", None)


    @property
    def relationships(self) -> dict:
        return self.data.get("relationships", {})


    @property
    def _parent_included(self) -> List[dict]:
        return self.get("_parent_included", [{}])


    def get_author(self):
        id = self.relationships.get("user", {}).get("data", {}).get("id", None)

        for raw_user in self._parent_included:
            if raw_user.get("id", None) == id and raw_user.get("type", None) == 'users':
                user = User(user=self.user, _fetched_data=dict(data=raw_user))

                return user

        return None


    def get_last_posted_user(self):
        id = self.relationships.get("lastPostedUser", {}).get("data", {}).get("id", None)

        for raw_user in self._parent_included:
            if raw_user.get("id", None) == id and raw_user.get("type", None) == 'users':
                user = User(user=self.user, _fetched_data=dict(data=raw_user))

                return user

        return None


    def get_first_post(self):
        id = self.relationships.get("firstPost", {}).get("data", {}).get("id", None)

        for raw_post in self._parent_included:
            if raw_post.get("id", None) == id and raw_post.get("type", None) == 'users':
                post = PostFromDiscussion(user=self.user, _fetched_data=dict(data=raw_post))

                return post

        return None


    def __restore_or_hide(self, hide: bool, force: bool=False):
        if hide:
            if self.isHidden and not force:
                raise FlarumError(f"Discussion {self.id} is already hidden. Use `force = True` to ignore this error.")

        else:
            if not self.isHidden and not force:
                raise FlarumError(f"Discussion {self.id} is already restored. Use `force = True` to ignore this error.")


        if not self.canHide and not force:
            raise FlarumError(f'You do not have permission to {"hide" if hide else "unhide"} this discussion ({self.id}). Use `force = True` to ignore this error.')

        return super().__restore_or_hide(hide=hide)


    def hide(self, force: bool=False):
        return self.__restore_or_hide(hide=True, force=force)


    def restore(self, force: bool=False):
        return self.__restore_or_hide(hide=False, force=force)
    unhide = restore


    def delete(self, force: bool=False):
        if not self.canDelete and not force:
            raise FlarumError(f'You do not have permission to delete this discussion ({self.id})')

        else:
            return super().delete()

# Circular:
from ...flarum.core.posts import PostFromBulk
class Discussion(DiscussionFromBulk):
    """
        A Flarum discussion.
    """

    def __init__(self, user: 'FlarumUser', _fetched_data: dict):
        self.user = user

        super().__init__(user=self.user, _fetched_data=_fetched_data)


    @property
    def included(self) -> List[dict]:
        return self.get("included", [{}])


    def get_posts(self):
        all_posts = list() # type: List[PostFromBulk]
        raw_posts = self.relationships.get("posts", {}).get("data", [{}]) # type: List[dict]

        for raw_post in raw_posts:
            if raw_post.get("type", None) == 'posts':
                id_to_find = raw_post.get("id", None)

                if id_to_find:
                    for possible_raw_post in self.included:
                        if (possible_raw_post.get("type", None) == 'posts') and (possible_raw_post.get("id", None) == id_to_find):
                            post = PostFromBulk(user=self.user, _fetched_data=dict(data=possible_raw_post, _parent_included=self.included))

                            all_posts.append(post)

        return all_posts


    def get_first_post(self):
        """
            The `Discussion` object does not have the first post's JSON data in it's own JSON. Because of Python's subclass inheritance, this
            function was included in `Discussion`, but it does not work!

            ### Alternative:

            ```python
            discussion = user.get_discussion_by_id(1)
            first_post = discussion.get_posts()[0]
            ```
        """

        raise NotImplementedError("`Discussion` does not have the first post data to fetch. Please, use the snippet in this function's docstring instead to obtain the first post.")



from typing import Dict, List


class ExtensionMixin:
    """
        A base class for mixing in custom classes (extensions) into another classes.

        ### Example extension code:
        ```python
        from pyflarum.extensions import ExtensionMixin
        from pyflarum.extensions.admin import AdminExtension
        from pyflarum.session import FlarumUser


        # Lowecase:
        AUTHOR = "yourname"
        NAME = "extensionname"
        ID = f"{AUTHOR}-{NAME}"

        # List of dependencies:
        SOFT_DEPENDENCIES = [AdminExtension] # uses methods from this extension, but can run without it
        HARD_DEPENCENDIES = []



        # I recommend to use the following naming pattern: `<YourExtensionName><ClassToMixin>Mixin`
        # Example:
        class ExampleFlarumUserMixin(FlarumUser):
            @property
            def __example(self):
                '''
                    Calling `FlarumUser(<...>).__example` would return this.
                '''

                return "Example"



        class ExampleExtension(ExtensionMixin):
            def get_dependencies(self):
                '''
                    Overwrite this method to make your own dependencies.

                    This should return the following `dict`:
                    ```python
                    {
                        "hard": [<class>, <class>, ...],
                        "soft": [<class>, <class>, ...]
                    }
                    ```

                    #### Hard-dependencies:
                    - Will raise an error when they're not found in the initialized `FlarumUser` object. It is impossible for the extension
                    to function without these.

                    #### Soft-dependencies:
                    - Will raise just a warning. It is possible for the extension to function without these, although with limitations
                    (such that some functions might be unavailable).
                '''

                return {
                    "soft": SOFT_DEPENDENCIES,
                    "hard": HARD_DEPENCENDIES
                }


            def mixin(self):
                '''
                    This extends the `FlarumUser` class with features from `ExampleMixin`
                '''

                super().mixin(self, FlarumUser, ExampleMixin)

        ```
    """

    def __init__(self):
        """
            Initializes the `ExtensionMixin` object.
        """

        self.name = "Unknown"
        self.author = "Unknown"
        self.id = "N/A"


    def get_dependencies(self) -> Dict[str, List[object]]:
        """
            Overwrite this method to make your own dependencies.

            This should return the following `dict`:
            ```python
            {
                "hard": [<class>, <class>, ...],
                "soft": [<class>, <class>, ...]
            }
            ```

            #### Hard-dependencies:
            - Will raise an error when they're not found in the initialized `FlarumUser` object. It is impossible for the extension
            to function without these.

            #### Soft-dependencies:
            - Will raise just a warning. It is possible for the extension to function without these, although with limitations
            (such that some functions might be unavailable).
        """

        return {
            "hard": [],
            "soft": []
        }


    def mixin(_, class_to_patch: object, class_to_mix_in: object):
        """
            A boilerplate function for mixing.
            
            This sets/overwrites attributes of `class_to_patch` to attributes of
            `class_to_mix_in` (monkey-patch).

            ### Example:
            ```python
            extension.mixin(myclass, pyflarum_class)
            ```
        """

        for prop, value in vars(class_to_mix_in).items():
            if not prop.startswith('__'):
                setattr(class_to_patch, f'{prop}', value)


# TODO: All included extensions
from .flarum.Flarum_Approval import *
from .flarum.FoF_BestAnswer import *


if __name__ == '__main__':
    print(
        ApprovalDiscussionFromNotificationMixin, ApprovalExtension,
        BestAnswerDiscussionMixin, BestAnswerExtension
    )



from typing import TYPE_CHECKING, Optional, List

# Avoid my greatest enemy in Python: circular import:
if TYPE_CHECKING:
    from ...session import FlarumUser

from ...error_handler import parse_request



class PreparedGroup(dict):
    def __init__(self, user: 'FlarumUser', nameSingular: str, namePlural: Optional[str]=None, color: Optional[str]=None, icon: Optional[str]=None, isHidden: bool=False):
        self.user = user

        self.nameSingular = nameSingular
        self.namePlural = (namePlural if namePlural else self.nameSingular)
        self.color = color
        self.icon = icon
        self.isHidden = isHidden
    

    @property
    def to_dict(self):
        data = {
            "data": {
                "type": "groups",
                "attributes": {
                    "nameSingular": self.nameSingular,
                    "namePlural": self.namePlural,
                    "color": self.color,
                    "icon": self.icon,
                    "isHidden": self.isHidden
                }
            }
        }

        return data


    def create(self):
        """
            Creates the group. Returns the created `Group`.
        """

        raw = self.user.session.post(self.user.api_urls['groups'], json=self.to_dict)
        json = parse_request(raw)

        return Group(user=self.user, _fetched_data=json)



class Groups(dict):
    """
        A data of multiple groups fetched from the API.
    """

    def __init__(self, user: 'FlarumUser', _fetched_data: dict):
        self.user = user

        super().__init__(_fetched_data)


    def __iter__(self):
        return iter(self.get_groups())


    @property
    def links(self) -> dict:
        return self.get("links", {})


    @property
    def first_link(self) -> Optional[str]:
        return self.links.get("first", None)


    @property
    def previous_link(self) -> Optional[str]:
        return self.links.get("prev", None)


    @property
    def next_link(self) -> Optional[str]:
        return self.links.get("next", None)


    @property
    def data(self) -> List[dict]:
        return self.get("data", [{}])


    @property
    def included(self) -> List[dict]:
        return self.get("included", [{}])


    def get_groups(self):
        """
            All groups from the `Groups` object.
        """

        all_groups = [] # type: List[Group]

        for raw_group in self.data:
            if raw_group.get("type", None) == 'groups':
                group = Group(user=self.user, _fetched_data=dict(data=raw_group))

                all_groups.append(group)

        return all_groups



class Group(Groups):
    """
        A Flarum group.
    """

    def __init__(self, user: 'FlarumUser', _fetched_data: dict):
        self.user = user

        super().__init__(user=self.user, _fetched_data=_fetched_data)


    @property
    def data(self) -> dict:
        return self.get("data", {})


    @property
    def type(self) -> Optional[str]:
        return self.data.get("type", None)


    @property
    def id(self) -> Optional[int]:
        raw = self.data.get("id", None)

        if raw:
            return int(raw)

    @property
    def attributes(self) -> dict:
        return self.data.get("attributes", {})


    @property
    def nameSingular(self) -> Optional[str]:
        return self.attributes.get("nameSingular", None)


    @property
    def namePlural(self) -> Optional[str]:
        return self.attributes.get("namePlural", None)


    @property
    def color(self) -> Optional[str]:
        return self.attributes.get("color", None)


    @property
    def icon(self) -> Optional[str]:
        return self.attributes.get("icon", None)


    @property
    def isHidden(self) -> bool:
        return bool(self.attributes.get("isHidden", 0))


    def edit(self, new_data: PreparedGroup):
        raw = self.user.session.patch(f"{self.user.api_urls['groups']}/{self.id}", json=new_data.to_dict)
        json = parse_request(raw)

        return Group(user=self.user, _fetched_data=json)


    def delete(self):
        raw = self.user.session.delete(f"{self.user.api_urls['groups']}/{self.id}")
        parse_request(raw)

        return True




from pathlib import Path

from .session import FlarumUser
from .error_handler import FlarumError

from .flarum.core.discussions import Discussion

from .flarum.core.filters import Filter


__description__ = "An unofficial Python package for manipulating with Flarum's API"
__author__      = "SKevo"
__copyright__   = "Copyright 2021, SKevo"
__credits__     = ["SKevo"]
__license__     = "GPLv3"
__version__     = "v1.0.6-beta"
__maintainer__  = "SKevo"
__email__       = "me@kevo.link"
__status__      = "4 - Beta"

# Default readme:
__readme__ = "(c) SKevo"

# Overwrite docstring, so pdoc can render it nicely
try:
    with open(f"{Path(__file__).parent.parent.absolute()}{Path('/README.md')}", 'r', encoding="UTF-8") as readme:
        __readme__ = readme.read()

except FileNotFoundError:
    pass


__doc__ = __readme__


if __name__ == "__main__":
    from typing import Union

    _: Union[
        FlarumUser,
        FlarumError,
        Discussion,
        Filter
    ]



from .. import ExtensionMixin

from ...flarum.core.users import UserFromBulk


AUTHOR = 'askvortsov'
NAME = 'moderator-warnings'
ID = f"{AUTHOR}-{NAME}"

SOFT_DEPENDENCIES = []
HARD_DEPENCENDIES = []


class ModeratorWarningsUserMixin(UserFromBulk):
    @property
    def canViewWarnings(self) -> bool:
        """
            Whether or not you can view the warnings of this user.
        """

        return self.attributes.get("canViewWarnings", False)


    @property
    def canManageWarnings(self) -> bool:
        return self.attributes.get("canManageWarnings", False)


    @property
    def canDeleteWarnings(self) -> bool:
        return self.attributes.get("canDeleteWarnings", False)


    @property
    def visibleWarningCount(self) -> int:
        return self.attributes.get("visibleWarningCount", 0)


class ModeratorWarningsExtension(ExtensionMixin):
    def get_dependencies(self):
        return {
            "soft": SOFT_DEPENDENCIES,
            "hard": HARD_DEPENCENDIES
        }


    def mixin(self):
        super().mixin(self, UserFromBulk, ModeratorWarningsUserMixin)



from .. import ExtensionMixin

from ...flarum.core.users import User, UserFromBulk, UserFromNotification
from ...error_handler import parse_request


AUTHOR = 'fof'
NAME = 'spamblock'
ID = f"{AUTHOR}-{NAME}"

SOFT_DEPENDENCIES = []
HARD_DEPENCENDIES = []



class SpamblockUserFromNotificationMixin(UserFromNotification):
    def spamblock(self) -> bool:
        raw = self.user.session.post(f"{self.user.api_urls['users']}/{self.id}/spamblock")
        parse_request(raw)

        return True



class SpamblockUserMixin(User, UserFromBulk, SpamblockUserFromNotificationMixin):
    @property
    def canSpamblock(self) -> bool:
        return self.attributes.get("canSpamblock", False)



class SpamblockExtension(ExtensionMixin):
    def get_dependencies(self):
        return {
            "soft": SOFT_DEPENDENCIES,
            "hard": HARD_DEPENCENDIES
        }


    def mixin(self):
        super().mixin(self, UserFromNotification, SpamblockUserFromNotificationMixin)
        super().mixin(self, UserFromBulk, SpamblockUserMixin)



from .. import ExtensionMixin

from ...flarum.core.discussions import DiscussionFromBulk


AUTHOR = 'flarum'
NAME = 'lock'
ID = f"{AUTHOR}-{NAME}"

SOFT_DEPENDENCIES = []
HARD_DEPENCENDIES = []


class LockDiscussionMixin(DiscussionFromBulk):
    @property
    def isLocked(self) -> bool:
        return self.attributes.get("isLocked", False)


    @property
    def canLock(self) -> bool:
        return self.attributes.get("canLock", False)



class StickyExtension(ExtensionMixin):
    def get_dependencies(self):
        return {
            "soft": SOFT_DEPENDENCIES,
            "hard": HARD_DEPENCENDIES
        }


    def mixin(self):
        super().mixin(self, DiscussionFromBulk, LockDiscussionMixin)



from typing import List, Optional, Union

from ...extensions import ExtensionMixin
from ...extensions.admin import AdminExtension, AdminFlarumUserMixin

from ...flarum.core.forum import Forum
from ...session import FlarumUser
from ...error_handler import parse_request


AUTHOR = 'malago'
NAME = 'achievements'
ID = f"{AUTHOR}-{NAME}"

SOFT_DEPENDENCIES = [AdminExtension]
HARD_DEPENCENDIES = []


class Achievement(dict):
    """
        An achievement (Malago's Achievements extension).
    """

    def __init__(self, user: 'FlarumUser', _fetched_data: dict):
        self.user = user

        super().__init__(_fetched_data)


    @property
    def data(self) -> dict:
        return self.get("data", {})


    @property
    def type(self) -> Optional[str]:
        return self.data.get("type", None)


    @property
    def attributes(self) -> dict:
        return self.data.get("attributes", {})


    @property
    def name(self) -> Optional[str]:
        return self.attributes.get("name", None)


    @property
    def description(self) -> Optional[str]:
        return self.attributes.get("description", None)


    @property
    def computation(self) -> Optional[str]:
        return self.attributes.get("computation", None)


    @property
    def points(self):
        raw = self.attributes.get("points", None)

        if raw:
            return int(raw)


    @property
    def icon(self) -> Optional[str]:
        return self.attributes.get("icon", None)


    @property
    def rectangle(self) -> Optional[str]:
        """
            No, I have no idea what this is either.
        """

        return self.attributes.get("rectangle", None)


    @property
    def active(self) -> Optional[bool]:
        raw = self.attributes.get("active", None)

        if raw:
            return bool(raw)


    @property
    def hidden(self) -> Optional[bool]:
        raw = self.attributes.get("hidden", None)

        if raw:
            return bool(raw)


    @property
    def new(self) -> Optional[str]:
        return self.attributes.get("new", None)


class AchievementsForumMixin(Forum):
    @property
    def show_achievements_in_post_footer(self) -> bool:
        return self.attributes.get("malago-achievements.show-post-footer", False)


    @property
    def show_achievements_in_user_card(self) -> bool:
        return self.attributes.get("malago-achievements.show-user-card", False)


class AchievementsAdminFlarumUserMixin(AdminFlarumUserMixin):
    def update_settings(self, show_achievement_list_in_each_post_footer: Optional[bool]=None, show_achievement_list_in_user_badge: Optional[bool]=None):
        post_data = {}


        if isinstance(show_achievement_list_in_each_post_footer, bool):
            post_data["malago-achievements.show-post-footer"] = show_achievement_list_in_each_post_footer

        if isinstance(show_achievement_list_in_user_badge, bool):
            post_data["malago-achievements.show-user-card"] = show_achievement_list_in_user_badge


        raw = self.session.post(f"{self.api_urls['settings']}", json=post_data)
        parse_request(raw)

        return True


    def create_achievement(self, name: str, description: str, computation: str, points: int, image_url_or_fa_icon: str, active: Union[bool, int]=1, hidden: Union[bool, int]=0):
        post_data = {
            "data": {
                "type": "achievements",
                "attributes": {
                    "name": name,
                    "description": description,
                    "computation": computation,
                    "points": points,
                    "image": image_url_or_fa_icon,
                    "rectangle": "0,0,,",
                    "active": int(active),
                    "hidden": int(hidden)
                }
            }
        }

        raw = self.session.post(f"{self.api_urls['base']}/achievements", json=post_data)
        json = parse_request(raw)

        return Achievement(user=self, _fetched_data=json)


    def get_all_achievements(self):
        raw = self.session.get(f"{self.api_urls['base']}/achievements")
        json = parse_request(raw)

        all_achievements = list() # type: List[Achievement]

        for raw_achievement in json['data']:
            achievement = Achievement(user=self, _fetched_data=dict(data=raw_achievement))
            all_achievements.append(achievement)

        return all_achievements



class AchievementsExtension(ExtensionMixin):
    def get_dependencies(self):
        return {
            "soft": SOFT_DEPENDENCIES,
            "hard": HARD_DEPENCENDIES
        }


    def mixin(self):
        super().mixin(self, AdminFlarumUserMixin, AchievementsAdminFlarumUserMixin)
        super().mixin(self, Forum, AchievementsForumMixin)
